/*****************************************************************************
 * RecommendationsService.kt
 *
 * Copyright © 2014-2018 VLC authors, VideoLAN and VideoLabs
 * Author: Geoffrey Métais
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 */
package org.videolan.vlc

import android.annotation.TargetApi
import android.app.IntentService
import android.app.Notification
import android.app.NotificationManager
import android.app.PendingIntent
import android.content.Context
import android.content.Intent
import android.os.Build
import android.support.v4.app.NotificationCompat
import android.support.v4.content.ContextCompat
import kotlinx.coroutines.experimental.CoroutineStart
import kotlinx.coroutines.experimental.launch
import kotlinx.coroutines.experimental.withContext
import org.videolan.medialibrary.media.MediaWrapper
import org.videolan.vlc.gui.helpers.BitmapUtil
import org.videolan.vlc.gui.video.VideoPlayerActivity
import org.videolan.vlc.util.Constants
import org.videolan.vlc.util.Util
import org.videolan.vlc.util.VLCIO
import org.videolan.vlc.util.getAppSystemService

private const val TAG = "VLC/RecommendationsService"
private const val MAX_RECOMMENDATIONS = 3

@TargetApi(Build.VERSION_CODES.JELLY_BEAN_MR1)
class RecommendationsService : IntentService("RecommendationService") {

    private lateinit var mNotificationManager: NotificationManager

    override fun onCreate() {
        super.onCreate()
        mNotificationManager = getAppSystemService(Context.NOTIFICATION_SERVICE) as NotificationManager
    }

    override fun onHandleIntent(intent: Intent?) {
        doRecommendations()
    }

    private fun buildRecommendation(mw: MediaWrapper?, id: Int, priority: Int) {
        if (mw == null) return
        // build the recommendation as a Notification object
        val notification = NotificationCompat.BigPictureStyle(
                NotificationCompat.Builder(this@RecommendationsService, "vlc_recommendations")
                        .setContentTitle(mw.title)
                        .setContentText(mw.description)
                        .setContentInfo(getString(R.string.app_name))
                        .setPriority(priority)
                        .setLocalOnly(true)
                        .setOngoing(true)
                        .setColor(ContextCompat.getColor(this, (application as VLCApplication).config.colorPrimary))
                        .setCategory(Notification.CATEGORY_RECOMMENDATION)
                        .setLargeIcon(BitmapUtil.getPicture(mw))
                        .setSmallIcon(R.drawable.app_icon)
                        .setContentIntent(buildPendingIntent(mw, id))
        ).build()
        // post the recommendation to the NotificationManager
        mNotificationManager.notify(id, notification)
    }

    private fun buildPendingIntent(mw: MediaWrapper, id: Int): PendingIntent {
        val intent = Intent(this@RecommendationsService, VideoPlayerActivity::class.java)
        intent.action = Constants.PLAY_FROM_VIDEOGRID
        intent.putExtra(Constants.PLAY_EXTRA_ITEM_LOCATION, mw.uri)
        intent.putExtra(Constants.PLAY_EXTRA_ITEM_TITLE, mw.title)
        intent.putExtra(Constants.PLAY_EXTRA_FROM_START, false)
        return PendingIntent.getActivity(this, id, intent, PendingIntent.FLAG_UPDATE_CURRENT)
    }

    private fun doRecommendations() = launch (start = CoroutineStart.UNDISPATCHED) {
        mNotificationManager.cancelAll()
        val videoList = withContext(VLCIO) { VLCApplication.getMLInstance().recentVideos }
        if (Util.isArrayEmpty(videoList)) return@launch
        for ((id, mediaWrapper) in videoList.withIndex()) {
            buildRecommendation(mediaWrapper, id, Notification.PRIORITY_DEFAULT)
            if (id == MAX_RECOMMENDATIONS) break
        }
    }
}
