package org.videolan.vlc.media

import android.app.Activity
import android.app.AlertDialog
import android.app.ProgressDialog
import android.content.Context
import android.content.DialogInterface
import android.net.Uri
import android.os.Handler
import android.os.Looper
import android.provider.MediaStore
import android.provider.OpenableColumns
import android.text.TextUtils
import android.util.Log
import android.view.WindowManager
import kotlinx.coroutines.experimental.CoroutineStart
import kotlinx.coroutines.experimental.android.UI
import kotlinx.coroutines.experimental.launch
import org.videolan.libvlc.util.AndroidUtil
import org.videolan.medialibrary.Tools
import org.videolan.medialibrary.media.MediaWrapper
import org.videolan.medialibrary.media.Playlist
import org.videolan.vlc.PlaybackService
import org.videolan.vlc.R
import org.videolan.vlc.VLCApplication
import org.videolan.vlc.util.*

private const val TAG = "VLC/MediaUtils"

object MediaUtils {

    private val subtitlesDownloader by lazy { SubtitlesDownloader() }

    @JvmOverloads
    fun getSubs(activity: Activity, mediaList: List<MediaWrapper>, cb: SubtitlesDownloader.Callback? = null) {
        subtitlesDownloader.downloadSubs(activity, mediaList, cb)
    }

    fun loadlastPlaylist(context: Context?, type: Int) {
        if (context == null) return
        DialogCallback(context, object : DialogCallback.Runnable {
            override fun run(service: PlaybackService, progress: ProgressDialog, errorDialog: AlertDialog.Builder) {
                service.loadLastPlaylist(type)
                progress.cancel();
            }
        })
    }

    @JvmOverloads
    fun getSubs(activity: Activity, media: MediaWrapper, cb: SubtitlesDownloader.Callback? = null) {
        getSubs(activity, mutableListOf(media), cb)
    }

    fun appendMedia(context: Context?, media: List<MediaWrapper>?) {
        if (media == null || context == null) return
        DialogCallback(context, object : DialogCallback.Runnable {
            override fun run(service: PlaybackService, progress: ProgressDialog, errorDialog: AlertDialog.Builder) {
                service.append(media)
                progress.cancel();
            }
        })
    }

    fun appendMedia(context: Context?, media: MediaWrapper?) {
        if (media == null || context == null) return
        DialogCallback(context, object : DialogCallback.Runnable {
            override fun run(service: PlaybackService, progress: ProgressDialog, errorDialog: AlertDialog.Builder) {
                service.append(media)
                progress.cancel();
            }
        })
    }

    fun appendMedia(context: Context, array: Array<MediaWrapper>) = appendMedia(context, array.asList())

    fun insertNext(context: Context?, media: Array<MediaWrapper>?) {
        if (media == null || context == null) return
        DialogCallback(context, object : DialogCallback.Runnable {
            override fun run(service: PlaybackService, progress: ProgressDialog, errorDialog: AlertDialog.Builder) {
                service.insertNext(media)
                progress.cancel();
            }
        })
    }

    fun insertNext(context: Context?, media: MediaWrapper?) {
        if (media == null || context == null) return
        DialogCallback(context, object : DialogCallback.Runnable {
            override fun run(service: PlaybackService, progress: ProgressDialog, errorDialog: AlertDialog.Builder) {
                service.insertNext(media)
                progress.cancel();
            }
        })
    }

    fun openMedia(context: Activity?, media: MediaWrapper?) {
        if (media == null || context == null) return
        AdLoader.loadFullscreenBanner(context) {
            if(Util.checkOnline(context)) {
                DialogCallback(context, object : DialogCallback.Runnable {
                    override fun run(service: PlaybackService, progress: ProgressDialog, errorDialog: AlertDialog.Builder) {
                        service.load(media, progress, errorDialog)
                    }
                })
            }
        }

    }

    fun openMediaNoUi(ctx: Context, id: Long) = launch(UI, CoroutineStart.UNDISPATCHED) {
        val media = ctx.getFromMl { getMedia(id) }
        openMediaNoUi(ctx, media)
    }

    fun openMediaNoUi(uri: Uri) = openMediaNoUi(VLCApplication.getAppContext(), MediaWrapper(uri))

    fun openMediaNoUi(context: Context?, media: MediaWrapper?) {
        if (media == null || context == null) return

        object : BaseCallBack(context) {
            override fun onConnected(service: PlaybackService) {
                service.load(media)
                client.disconnect()
            }
        }
    }

    fun openArray(context: Context, array: Array<MediaWrapper>, position: Int) = openList(context, array.toList(), position)

    @JvmOverloads
    fun openList(context: Context?, list: List<MediaWrapper>, position: Int, shuffle: Boolean = false) {
        if (Util.isListEmpty(list) || context == null) return
        AdLoader.loadFullscreenBanner(context) {
            DialogCallback(context, object : DialogCallback.Runnable {
                override fun run(service: PlaybackService, progress: ProgressDialog, errorDialog: AlertDialog.Builder) {
                    service.load(list, position, null, null)
                    if (shuffle && !service.isShuffling) service.shuffle()
                    progress.cancel();
                }
            })
        }

    }

    fun openUri(context: Context?, uri: Uri?) {
        if (uri == null || context == null) return

        AdLoader.loadFullscreenBanner(context) {
            DialogCallback(context, object : DialogCallback.Runnable {
                override fun run(service: PlaybackService, progress: ProgressDialog, errorDialog: AlertDialog.Builder) {
                    service.loadUri(uri)
                    progress.cancel();
                }
            })
        }
    }

    fun openStream(context: Context?, uri: String?) {
        if (uri == null || context == null) return
        DialogCallback(context, object : DialogCallback.Runnable {
            override fun run(service: PlaybackService, progress: ProgressDialog, errorDialog: AlertDialog.Builder) {
                service.loadLocation(uri)
            }
        })
    }

    fun getMediaArtist(ctx: Context, media: MediaWrapper?) = media?.artist ?: getMediaString(ctx, R.string.unknown_artist)

    fun getMediaReferenceArtist(ctx: Context, media: MediaWrapper?) = media?.artist ?: getMediaString(ctx, R.string.unknown_artist)

    fun getMediaAlbumArtist(ctx: Context, media: MediaWrapper?) = media?.albumArtist ?: getMediaString(ctx, R.string.unknown_artist)

    fun getMediaAlbum(ctx: Context, media: MediaWrapper?) = media?.album ?: getMediaString(ctx, R.string.unknown_album)

    fun getMediaGenre(ctx: Context, media: MediaWrapper?) = media?.genre ?: getMediaString(ctx, R.string.unknown_genre)

    fun getMediaSubtitle(media: MediaWrapper): String? {
        var subtitle = media.nowPlaying ?: media.artist
        if (media.length > 0L) {
            subtitle = if (TextUtils.isEmpty(subtitle)) Tools.millisToString(media.length)
            else "$subtitle  -  ${Tools.millisToString(media.length)}"
        }
        return subtitle
    }

    fun getMediaTitle(mediaWrapper: MediaWrapper) = mediaWrapper.title ?: FileUtils.getFileNameFromPath(mediaWrapper.location)!!

    fun getContentMediaUri(data: Uri): Uri {
        VLCApplication.getAppContext().contentResolver.query(data,
                arrayOf(MediaStore.Video.Media.DATA), null, null, null)?.use {
            val columnIndex = it.getColumnIndexOrThrow(MediaStore.Video.Media.DATA)
            if (it.moveToFirst()) return AndroidUtil.PathToUri(it.getString(columnIndex)) ?: data
        }
        return data
    }

    private fun getMediaString(ctx: Context?, id: Int): String {
        return if (ctx != null) ctx.resources.getString(id)
        else when (id) {
            R.string.unknown_artist -> "Unknown Artist"
            R.string.unknown_album -> "Unknown Album"
            R.string.unknown_genre -> "Unknown Genre"
            else -> ""
        }
    }

    private abstract class BaseCallBack : PlaybackService.Client.Callback {
        protected lateinit var client: PlaybackService.Client

        internal constructor(context: Context) {
            client = PlaybackService.Client(context, this)
            client.connect()
        }

        protected constructor()


        override fun onDisconnected() {}
    }

    private class DialogCallback (context: Context, private val runnable: Runnable) : BaseCallBack() {
        private lateinit var dialog: ProgressDialog
        private val handler = Handler(Looper.getMainLooper())
        protected var errorDialog: AlertDialog.Builder

        internal interface Runnable {
            fun run(service: PlaybackService, progress: ProgressDialog, errorDialog: AlertDialog.Builder)
        }

        init {
            client = PlaybackService.Client(context, this)

            this.errorDialog = AlertDialog.Builder(context).setPositiveButton(android.R.string.ok) { dialogInterface, i -> dialogInterface.cancel() }


            handler.postDelayed({
                dialog = try{
                    ProgressDialog.show(
                            context,
                            context.applicationContext.getString(R.string.loading) + "…",
                            context.applicationContext.getString(R.string.please_wait), true)
                }catch (e: WindowManager.BadTokenException){
                    return@postDelayed
                }
                dialog.setCancelable(true)
                dialog.setOnCancelListener(object : DialogInterface.OnCancelListener {
                    override fun onCancel(dialog: DialogInterface) {
                        synchronized(this) {
                            client.disconnect()
                        }
                    }
                })

                Log.d("LoadingDialog", "show")
                dialog.show()

                synchronized(this) {

                    Log.d("LoadingDialog", "connect")
                    try {
                        client.connect()
                    } catch (e: Exception) {
                        Log.d("LoadingDialog", "exception cancel")

                        dialog.cancel()
                        throw e
                    }
                }

            }, 300)



        }

        override fun onConnected(service: PlaybackService) {
            synchronized(this) {
                runnable.run(service, dialog, errorDialog)
            }
        }

        override fun onDisconnected() {
            Log.d("LoadingDialog", "disconnected");
            runOnMainThread(Runnable {
                errorDialog.setMessage("An error occured during the loading. Please try again.").show()
            })
            if (this::dialog.isInitialized) dialog.dismiss()
        }
    }

    fun retrieveMediaTitle(mw: MediaWrapper) = try {
        VLCApplication.getAppContext().contentResolver.query(mw.uri, null, null, null, null)?.use {
            val nameIndex = it.getColumnIndex(OpenableColumns.DISPLAY_NAME)
            if (nameIndex > -1 && it.count > 0) {
                it.moveToFirst()
                if (!it.isNull(nameIndex)) mw.title = it.getString(nameIndex)
            }
        }
    } catch (ignored: UnsupportedOperationException) {}

    fun deletePlaylist(playlist: Playlist) = runBackground(Runnable { playlist.delete() })
}
